from datetime import datetime

from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Sequence
from sqlalchemy.engine import Engine
from sqlalchemy.schema import Column
from sqlalchemy.schema import Table


def _primary_key_column(
        engine: Engine, column_name: str, table_name: str
) -> Column:
    if engine.url.get_backend_name() == 'duckdb':
        # DuckDB needs a little help with primary keys:
        # https://github.com/Mause/duckdb_engine#auto-incrementing-id-columns
        sensor_id_seq = Sequence(f'{table_name}_id_seq')
        id_column = Column(column_name,
                           Integer,
                           sensor_id_seq,
                           server_default=sensor_id_seq.next_value(),
                           primary_key=True)
    else:
        id_column = Column(column_name, Integer, primary_key=True)

    return id_column


def generate_scorpion_testdb(engine: Engine):
    """Create a test DB from Table 1 of the Scorpion paper (Wu & Madden,
    VLDB 2013).

    Rather than make sensor_id a foreign key, we make it a character
    type to test a wider variety of column types.
    """
    metadata = MetaData(engine)
    sensor_readings = Table(
        'sensor_readings',
        metadata,
        _primary_key_column(engine, 'id', 'sensor_readings'),
        Column('created_at', DateTime, nullable=False),
        Column('sensor_id', String, nullable=False),
        Column('voltage', Float, nullable=False),
        Column('humidity', Float, nullable=False),
        Column('temperature', Float, nullable=False),
    )
    metadata.create_all(engine)
    values = [
        (datetime(2021, 5, 5, 11), '1', 2.64, 0.4, 34),
        (datetime(2021, 5, 5, 11), '2', 2.65, 0.5, 35),
        (datetime(2021, 5, 5, 11), '3', 2.63, 0.4, 35),
        (datetime(2021, 5, 5, 12), '1', 2.7, 0.3, 35),
        (datetime(2021, 5, 5, 12), '2', 2.7, 0.5, 35),
        (datetime(2021, 5, 5, 12), '3', 2.3, 0.4, 100),
        (datetime(2021, 5, 5, 13), '1', 2.7, 0.3, 35),
        (datetime(2021, 5, 5, 13), '2', 2.7, 0.5, 35),
        (datetime(2021, 5, 5, 13), '3', 2.3, 0.5, 80),
    ]
    with engine.connect() as conn:
        conn.execute(sensor_readings.insert(), [
            dict(zip(
                ('created_at', 'sensor_id',
                 'voltage', 'humidity', 'temperature'),
                value))
            for value in values
        ])


def generate_synthetic_testdb(engine: Engine):
    """Create a synthetic database with several data types and
    distributions.
    For various data types (datetime, integer, string, float), we
    generate distributions that are:
    * Unique overall (every value is different)
    * The same overall (every value is the same)
    * Unique within a bucket (but repeats across buckets).
    """
    metadata = MetaData(engine)
    synthetic_data = Table(
        'synthetic_data',
        metadata,
        _primary_key_column(engine, 'id', 'synthetic_data'),
        Column('same_datetime', DateTime, nullable=False),
        Column('unique_datetime', DateTime, nullable=False),
        Column('bucket_unique_datetime', DateTime, nullable=False),
        Column('same_string', String, nullable=False),
        Column('unique_string', String, nullable=False),
        Column('bucket_unique_string', String, nullable=False),
        Column('same_float', Float, nullable=False),
        Column('unique_float', Float, nullable=False),
        Column('bucket_unique_float', Float, nullable=False),
        Column('same_int', Integer, nullable=False),
        Column('unique_int', Integer, nullable=False),
        Column('bucket_unique_int', Integer, nullable=False),
    )
    metadata.create_all(engine)
    values = []
    for bucket in range(1, 20):
        for row in range(1, 10):
            values.append((
                datetime(2021, 5, 5, 11),
                datetime(2021, 5, bucket, row),
                datetime(2021, 5, 5, row),
                'hi',
                f'{bucket}-{row}',
                f'{row}',
                1.1,
                bucket * 1.0 + (row / 100.0),
                row * 1.0,
                1,
                bucket * 1000 + row,
                row
            ))
    with engine.connect() as conn:
        conn.execute(synthetic_data.insert(), [
            dict(zip(
                ('same_datetime', 'unique_datetime', 'bucket_unique_datetime',
                 'same_string', 'unique_string', 'bucket_unique_string',
                 'same_float', 'unique_float', 'bucket_unique_float',
                 'same_int', 'unique_int', 'bucket_unique_int'), value))
            for value in values
        ])



from datetime import datetime
from typing import Union

from sqlalchemy.engine import Engine


def engine_based_datetime(engine: Engine, string: str) -> Union[str, datetime]:
    """
    If the engine is sqlite (which represents datetimes as strings), we
    return a string representation of a datetime. If it's anything else,
    we return a Pythonic version.
    """
    if engine.url.get_backend_name() == 'sqlite':
        return string

    return datetime.strptime(string, '%Y-%m-%d %H:%M:%S.%f')



import sqlalchemy

from collections import defaultdict
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Set

from datools.models import Column
from datools.models import Table


RANGE_VALUED_TYPES = {
    sqlalchemy.sql.sqltypes.BigInteger,
    sqlalchemy.sql.sqltypes.Date,
    sqlalchemy.sql.sqltypes.DateTime,
    sqlalchemy.sql.sqltypes.Float,
    sqlalchemy.sql.sqltypes.Integer,
    sqlalchemy.sql.sqltypes.Interval,
    sqlalchemy.sql.sqltypes.Numeric,
    sqlalchemy.sql.sqltypes.SmallInteger,
    sqlalchemy.sql.sqltypes.Time,
}
SET_VALUED_TYPES = {
    sqlalchemy.sql.sqltypes.Boolean,
    sqlalchemy.sql.sqltypes.Enum,
    sqlalchemy.sql.sqltypes.Integer,
    sqlalchemy.sql.sqltypes.SmallInteger,
    sqlalchemy.sql.sqltypes.String,
    sqlalchemy.sql.sqltypes.Text,
    sqlalchemy.sql.sqltypes.Unicode,
    sqlalchemy.sql.sqltypes.UnicodeText,
}


class ColumnStatistics:
    pass


@dataclass
class SetValuedStatistics(ColumnStatistics):
    distinct_values: int
    most_common_values: list

    def __repr__(self):
        return (
            f'SetValuedStatistics(distinct_values: {self.distinct_values}, '
            f'most_common_values: {self.most_common_values})')

    def __eq__(self, other):
        return (self.distinct_values == other.distinct_values
                and (set(self.most_common_values) ==
                     set(other.most_common_values)))


@dataclass
class RangeValuedStatistics(ColumnStatistics):
    bucket_minimums: List[Any]

    def __repr__(self):
        return (
            f'RangeValuedStatistics(bucket_minimums: '
            f'{[minimum for minimum in self.bucket_minimums]})')


def set_valued_statistics(
        engine: sqlalchemy.engine.Engine,
        query: str,
        columns: Set[Column],
        num_most_common_values: int = 100
) -> List[Tuple[Column, SetValuedStatistics]]:
    clauses: List[str] = []
    for column in columns:
        clauses.append(
            f'COUNT(DISTINCT {column.name})'
            f'AS {column.name}_distinct_values')
    results = engine.execute(
        f'WITH query AS ( '
        f'{query}'
        f')'
        f'SELECT {", ".join(clauses)} FROM query')
    count_statistics = list(results)[0]
    results.close()
    statistics: List[Tuple[Column, SetValuedStatistics]] = []
    for column in columns:
        results = engine.execute(
                f'WITH query AS ( '
                f'{query}'
                f')'
                f'SELECT {column.name}, COUNT(*) AS num_rows '
                f'FROM query '
                f'GROUP BY {column.name} '
                f'ORDER BY num_rows DESC, {column.name} '
                f'LIMIT {num_most_common_values}')
        values = [row[column.name] for row in results]
        results.close()
        statistics.append((
            column,
            SetValuedStatistics(
                count_statistics[f'{column.name}_distinct_values'],
                values)))
    return statistics


def range_valued_statistics(
        engine: sqlalchemy.engine.Engine,
        query: str,
        columns: Set[Column],
        num_buckets: int = 3
) -> List[Tuple[Column, RangeValuedStatistics]]:
    bucket_clauses: List[str] = []
    first_clauses: List[str] = []
    for column in columns:
        bucket_clauses.append(
            f'{column.name}, '
            f'NTILE({num_buckets}) OVER (ORDER BY {column.name} ASC) '
            f'AS {column.name}_bucket')
        first_clauses.append(
            f'first_value({column.name}) OVER '
            f'(PARTITION BY {column.name}_bucket '
            f' ORDER BY {column.name} ASC '
            f' RANGE BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) '
            f'AS {column.name}_bucket_value')

    statistics: List[Tuple[Column, RangeValuedStatistics]] = []
    if bucket_clauses:
        results = engine.execute(
            f'WITH query AS ( '
            f'{query}'
            f'), '
            f'buckets AS ( '
            f'SELECT {", ".join(bucket_clauses)} '
            f'FROM query '
            f')'
            f'SELECT {", ".join(first_clauses)} '
            f'FROM buckets ')
        # The output of the window function is the number of rows in the
        # input rather than the number of range buckets. Because we have a
        # single query across multiple columns, we can't `GROUP BY
        # column_name_bucket` and get the `MIN` NTILE value for each
        # bucket. Since sqlite doesn't support CUBE/ROLLUP/GROUPING SETS,
        # we do this grouping manually by creating a `set` of the
        # `first_value` for each bucket.
        column_values: Dict[Column, Set[Any]] = defaultdict(set)
        for row in results:
            for column in columns:
                value = row[f'{column.name}_bucket_value']
                # Some engines (e.g., SQLite) happily store strings in
                # numeric columns, so we have to be a bit defensive of
                # the values we get back.
                if column is not None and not value == '':
                    column_values[column].add(value)
        results.close()
        for column in columns:
            statistics.append((
                column,
                RangeValuedStatistics(sorted(column_values[column]))))
    return statistics


def column_statistics(
        engine: sqlalchemy.engine.Engine,
        table: Table,
        columns_to_ignore: Set[Column],
) -> Dict[Column, List[ColumnStatistics]]:
    metadata = sqlalchemy.MetaData()
    table_metadata = sqlalchemy.Table(
        table.name, metadata, autoload_with=engine)
    candidate_columns = [column for column in table_metadata.columns
                         if Column(column.name) not in columns_to_ignore]
    statistics: Dict[Column, List[ColumnStatistics]] = defaultdict(list)
    for column, set_statistic in set_valued_statistics(
            engine,
            f'SELECT * FROM {table.name}',
            {Column(column.name) for column in candidate_columns if
             type(column.type.as_generic()) in SET_VALUED_TYPES}):
        statistics[column].append(set_statistic)
    for column, range_statistic in range_valued_statistics(
            engine,
            f'SELECT * FROM {table.name}',
            {Column(column.name) for column in candidate_columns if
             type(column.type.as_generic()) in RANGE_VALUED_TYPES}):
        statistics[column].append(range_statistic)
    return statistics



import pytest

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine


DB_TYPE_TO_ENGINE_STRING = {
    'sqlite': 'sqlite://',
    'duckdb': 'duckdb:///:memory:'
}


def pytest_addoption(parser):
    parser.addoption('--db-type',
                     required=True,
                     choices=DB_TYPE_TO_ENGINE_STRING.keys())


@pytest.fixture(scope='function')
def db_engine(pytestconfig) -> Engine:
    engine_string = DB_TYPE_TO_ENGINE_STRING[pytestconfig.getoption('db_type')]
    return create_engine(engine_string)



#!/usr/bin/env python

from pytest import approx
from sqlalchemy.engine import Engine

from datools.models import Column
from datools.models import Constant
from datools.models import Explanation
from datools.models import Operator
from datools.models import Predicate
from datools.explanations import diff
from .fixtures import generate_scorpion_testdb


def test_diff(db_engine: Engine):
    generate_scorpion_testdb(db_engine)
    candidates = diff(
        db_engine,
        'SELECT * FROM sensor_readings WHERE temperature > 50',
        'SELECT * FROM sensor_readings WHERE temperature <= 50',
        {Column('created_at'), Column('sensor_id'), Column('voltage'),
         Column('humidity')},
        {Column('voltage'), Column('humidity')},
        0.05,
        2.0,
        1)
    assert(candidates == [
        Explanation(
            (Predicate(
                Column('voltage'), Operator.EQUALS, Constant(approx(2.3))), ),
            risk_ratio=9.0),
        Explanation(
            (Predicate(Column('sensor_id'), Operator.EQUALS, Constant('3')), ),
            risk_ratio=5 + (1.0 / 3))])



from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Tuple


AggregateFunction = Enum(
    'AggregateFunction',
    'SUM COUNT AVERAGE')


Operator = Enum(
    'Operator',
    'EQUALS NOT_EQUALS GT LT GTEQ LTEQ')


OPERATOR_TO_SQL = {
    Operator.EQUALS: '=',
    Operator.NOT_EQUALS: '<>',
    Operator.GT: '>',
    Operator.LT: '<',
    Operator.GTEQ: '>=',
    Operator.LTEQ: '<=',
}


@dataclass
class Column:
    name: str

    def __hash__(self):
        return self.name.__hash__()

    def __eq__(self, other):
        return self.name.__eq__(other.name)


@dataclass
class Constant:
    value: Any


@dataclass
class Aggregate:
    function: AggregateFunction
    column: Column


@dataclass
class Predicate:
    left: Column
    operator: Operator
    right: Constant

    def to_sql(self):
        return (f'{self.left.name} '
                f'{OPERATOR_TO_SQL[self.operator]} {self.right.value}')

    def __repr__(self):
        return f'Predicate({self.to_sql()})'


@dataclass
class Table:
    name: str


@dataclass
class Explanation:
    predicates: Tuple[Predicate, ...]
    risk_ratio: float



import sqlalchemy
from tabulate import tabulate

from typing import Dict
from typing import Tuple

from datools.models import Column


GROUP_COLUMNS_KEY = '%%%GROUP_COLUMNS%%%'
GROUPING_SETS_KEY = '%%%GROUPING_SETS%%%'
GROUPING_ID_KEY = '%%%GROUPING_ID%%%'
INDENT = '    '


def query_columns(
        engine: sqlalchemy.engine.Engine, query: str
) -> Tuple[str, ...]:
    results = engine.execute(query)
    columns = tuple(column[0] for column in results.cursor.description)
    results.close()
    return columns


def query_results_pretty_print(
        engine: sqlalchemy.engine.Engine, query: str, label: str = None
) -> None:
    if label:
        print(f'*** {label} ***')
    result = engine.execute(query)
    all_rows = (dict(row) for row in result)
    print(tabulate(all_rows, headers='keys', tablefmt='psql'))
    result.close()


def query_rows(engine: sqlalchemy.engine.Engine, query: str) -> int:
    count_query = (
        f'WITH query AS ({query}) '
        f'SELECT COUNT(*) AS num_rows FROM query')
    results = engine.execute(count_query)
    rows = results.first().num_rows
    results.close()
    return rows


def grouping_sets_query(
        engine: sqlalchemy.engine.Engine,
        query: str,
        sets: Tuple[Tuple[Column, ...], ...],
        group_columns_key: str = GROUP_COLUMNS_KEY,
        grouping_sets_key: str = GROUPING_SETS_KEY,
        grouping_id_key: str = GROUPING_ID_KEY
) -> Tuple[str, Dict[int, Tuple[Column, ...]]]:
    """Takes a `query` like

    SELECT
        {grouping_id_key} AS grouping_id,
        {group_columns_key},
        AGGREGATE1(args), AGGREGATE2(args)
    ...
    GROUP BY {grouping_sets_key}
    ...

    and returns a tuple with
      * a grouping sets query for the grouping sets in `sets`, and
      * a dictionary mapping set IDs to grouping set columns.

    If the database that `engine` is connected to natively supports
    grouping sets, utilize the standard SQL syntax for them. If it doesn't,
    implement the query by capturing the UNION ALL output of multiple
    GROUP BY subqueries.

    The `grouping_id` will be set to a unique value for each grouping set,
    and will be deterministic on the iteration order of `sets`.
    """
    if engine.url.get_backend_name() == 'postgresql':
        # TODO(marcua): Implement grouping sets using SQL syntax in
        # databases that support it.
        pass
    column_indices: Dict[Column, int] = {}
    for grouping_set in sets:
        for column in grouping_set:
            index = column_indices.get(column)
            if index is None:
                column_indices[column] = len(column_indices)

    queries = []
    set_index: Dict[int, Tuple[Column, ...]] = {}
    for set_id, grouping_set in enumerate(sets):
        set_index[set_id] = grouping_set
        group_columns = [f'NULL AS {column.name}'
                         for column in column_indices.keys()]
        for column in grouping_set:
            group_columns[column_indices[column]] = column.name
        grouping_set_names = tuple(column.name for column in grouping_set)
        queries.append(
            query
            .replace(grouping_id_key, str(set_id))
            .replace(group_columns_key, ', '.join(group_columns))
            .replace(grouping_sets_key, ', '.join(grouping_set_names)))
    return '\nUNION ALL\n'.join(queries), set_index



import sqlalchemy

from collections import defaultdict
from math import floor
from textwrap import dedent
from textwrap import indent
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from datools.errors import DatoolsError
from datools.models import Column
from datools.models import Constant
from datools.models import Explanation
from datools.models import Operator
from datools.models import Predicate
from datools.sqlalchemy_utils import GROUP_COLUMNS_KEY
from datools.sqlalchemy_utils import GROUPING_ID_KEY
from datools.sqlalchemy_utils import GROUPING_SETS_KEY
from datools.sqlalchemy_utils import INDENT
from datools.sqlalchemy_utils import grouping_sets_query
from datools.sqlalchemy_utils import query_columns
from datools.sqlalchemy_utils import query_rows
from datools.table_statistics import range_valued_statistics
from datools.table_statistics import RangeValuedStatistics


NUM_RANGE_BUCKETS = 15


def _rewrite_query_with_ranges_as_buckets(
        query: str,
        range_statistics: List[Tuple[Column, RangeValuedStatistics]]
) -> Tuple[str, Dict[Column, List[Tuple[Predicate, ...]]]]:
    # For each of the columns we've got range predicates on, create a
    # proxy column for the bucketed range values.
    bucket_predicates: Dict[Column, List[Tuple[Predicate, ...]]] = defaultdict(
        list)
    for column, statistic in range_statistics:
        # If the bucket minimums in the statistics are (a, b, c),
        # we want to generate the following ranges:
        # * < b [we leave out a in case the control relation
        #        has smaller values than the test relation]
        # * >= b, < c [values in the middle bucket]
        # * >= c [the largest bucket has a minimum of c]
        pairs = zip([None] + statistic.bucket_minimums[1:],
                    statistic.bucket_minimums[1:] + [None])
        for first, second in pairs:
            bucket_column = Column(f'{column.name}__bucket')
            first_predicate = Predicate(
                column, Operator.GTEQ, Constant(first))
            second_predicate = Predicate(column, Operator.LT, Constant(second))
            if first is None and second is not None:
                bucket_predicates[bucket_column].append((second_predicate, ))
            elif first is not None and second is None:
                bucket_predicates[bucket_column].append((first_predicate, ))
            # Don't allow a 1-bucket degenerate case (`first` and `second`
            # are empty).
            elif first is not None and second is not None:
                bucket_predicates[bucket_column].append(
                    (first_predicate, second_predicate))

    # Create a CASE statement for each bucket column that converts
    # each range within that column into a single bucketed value.
    cases = []
    for column, column_predicates in bucket_predicates.items():
        whens = []
        for index, predicate_group in enumerate(column_predicates):
            clause = ' AND '.join(predicate.to_sql()
                                  for predicate in predicate_group)
            whens.append(f'WHEN {clause} THEN {index}')
        when_lines = indent(f'\n'.join(whens), 4 * INDENT)
        cases.append(f'CASE\n{when_lines}\nEND AS {column.name}')

    # Generate SQL.
    case_lines = ',\n'.join(cases)
    return dedent(
        f'''
        WITH original_query AS (
            {query}
        )
        SELECT
            original_query.*{',' if case_lines else ''}
            {case_lines}
        FROM original_query
        '''), bucket_predicates


def _explanation_counts_query(
        engine: sqlalchemy.engine.Engine,
        relation: str,
        on_columns: Set[Column],
        min_support_rows: int = None
) -> Tuple[str, Dict[int, Tuple[Column, ...]]]:
    explanation_query = dedent(
        f'''
        SELECT
            {GROUPING_ID_KEY} AS grouping_id,
            {GROUP_COLUMNS_KEY},
            1.0 * COUNT(*) AS explanation_size
        FROM query
        GROUP BY {GROUPING_SETS_KEY}
        ''')
    if min_support_rows is not None:
        explanation_query += f'HAVING explanation_size > {min_support_rows}\n'

    group_explanations_query, grouping_set_index = grouping_sets_query(
        engine,
        explanation_query,
        tuple((column, ) for column in on_columns))
    return dedent(
        f'''
        WITH query AS (
            {relation}
        )
        {group_explanations_query}
        '''), grouping_set_index


def _diff_query(
        test_explanations_query: str,
        control_explanations_query: str,
        num_test_rows: float,
        num_control_rows: float,
        on_columns: Set[Column],
        min_risk_ratio: float
) -> str:
    join_conditions = (
        ['test.grouping_id = control.grouping_id']
        + [f'((test.{column.name} = control.{column.name}) '
           f' OR ((test.{column.name} IS NULL) '
           f'     AND (control.{column.name} IS NULL)))'
           for column in on_columns])
    join_statement = ' AND '.join(join_conditions)

    # TODO(marcua): Consult with someone better at statistics on how
    # to avoid division by 0 in the risk ratio when a group encompases
    # the entire relation. For now, make the relation size one larger
    # than it actually is.
    adjusted_test_rows = num_test_rows + 1
    adjusted_control_rows = num_control_rows + 1

    diff_query = dedent(
        f'''
        WITH
        test AS (
            {indent(test_explanations_query, 3 * INDENT)}
        ),
        control AS (
            {indent(control_explanations_query, 3 * INDENT)}
        )
        SELECT
            test.grouping_id,
            {', '.join(f'test.{column.name}' for column in on_columns)},
            test.explanation_size AS test_explanation_size,
            control.explanation_size AS control_explanation_size,
            (test.explanation_size / (test.explanation_size
                                      + COALESCE(control.explanation_size, 0)))
            /
            (({adjusted_test_rows} - test.explanation_size)
             / (({adjusted_test_rows} - test.explanation_size)
                + ({adjusted_control_rows}
                   - COALESCE(control.explanation_size, 0)))
            ) AS risk_ratio
        FROM test
        LEFT JOIN control ON {join_statement}
        WHERE risk_ratio > {min_risk_ratio}
        ORDER BY risk_ratio DESC
        ''')
    return diff_query


def diff(
        engine: sqlalchemy.engine.Engine,
        test_relation: str,
        control_relation: str,
        on_column_values: Set[Column],
        on_column_ranges: Set[Column],
        min_support: float,
        min_risk_ratio: float,
        max_order: int
) -> List[Explanation]:
    """
    Generates candidate explanations for why records are more likely to appear
    in `test_relation` than `control_relation`. For example, if you provided
    a test set of customers who are more likely to churn and a control set of
    customers who are less likely to churn, the algorithm might list predicates
    related to business category or degree of service.

    This is an implementatin of "DIFF: A Relational Interface
    for Large-Scale Data Explanation"
    by Firas Abuzaid, Peter Kraft, Sahaana Suri, Edward Gan,
    Eric Xu, Atul Shenoy, Asvin Ananthanarayan, John Sheu,
    Erik Meijer, Xi Wu, Jeff Naughton, Peter Bailis, and Matei Zaharia.
    http://www.bailis.org/papers/diff-vldb2019.pdf

    Figure 1 of that paper contains a declarative description of the
    implemented algorithm.

    :param engine: A SQLAlchemy engine.
    :param test_relation: A SQL query resulting in rows whose presence
                          you would like to explain.
    :param control_relation: A SQL query (with the same schema
                             as `test_relation`) that contains a control
                             group of rows that are different from/a
                             baseline for `test_relation`.
    :param on_column_values: A set of columns that might contain a specific
                       value that explains the difference between
                       test and control. Example: the column `sensor_id`
                       might contain a specific sensor's ID that is the
    :param on_column_ranges: A set of columns that might contain a range of
                       values that explains the difference between
                       test and control. Example: the column `timestamp`
                       might contain a date/time range that is the
                       source of outlier readings.
    :param min_support: A value in the range [0, 1] with the minimum fraction
                        of the test set that an explanation should contain.
    :param min_risk_ratio: A positive floating point value with the minimum
                           risk ratio of an explanation. A high risk ratio
                           means an explanation is more likely to describe a
                           row in `test_relation` than `control_relation`.
    :param max_order: The largest number of columns on which to consider
                      an explanation. For example, a 2-column explanation could
                      be that (category='dog walker' AND signup_day='sunday').
                      The current implementation only supports 1-column
                      explanations.
    """
    if max_order != 1:
        raise DatoolsError('Only one-column predicates are supported for now')

    # Get all column names from test_relation and control_relation,
    # ensure they are the same.
    # TODO(marcua): compare types.
    test_column_names = query_columns(engine, test_relation)
    control_column_names = query_columns(engine, control_relation)
    if test_column_names != control_column_names:
        raise DatoolsError(
            'test_relation and control_relation have different schemas')

    # Ensure on_columns are a subset of the test/control columns.
    on_column_names = ({column.name for column in on_column_values} |
                       {column.name for column in on_column_ranges})
    if on_column_names - set(test_column_names):
        raise DatoolsError('on_columns is not a subset of test_relation')

    # Get size of test_relation, control_relation.
    num_test_rows = 1.0 * query_rows(engine, test_relation)
    num_control_rows = 1.0 * query_rows(engine, control_relation)
    min_support_rows = floor(num_test_rows * min_support)

    # Transform ranges in on_column_ranges into bucket IDs.
    range_statistics = range_valued_statistics(
        engine, test_relation, on_column_ranges,
        num_buckets=NUM_RANGE_BUCKETS)
    rewritten_test_relation, test_bucket_predicates = (
        _rewrite_query_with_ranges_as_buckets(
            test_relation, range_statistics))
    rewritten_control_relation, control_bucket_predicates = (
        _rewrite_query_with_ranges_as_buckets(
            control_relation, range_statistics))

    # GROUP BY all test_relation columns, remove ones with a size less
    # than min_support_rows.
    on_columns = (on_column_values |
                  {column for column in test_bucket_predicates.keys()})
    test_explanations_query, grouping_set_index = _explanation_counts_query(
        engine, rewritten_test_relation, on_columns, min_support_rows)
    control_explanations_query, _ = _explanation_counts_query(
        engine, rewritten_control_relation, on_columns, min_support_rows=None)
    diff_query = _diff_query(
        test_explanations_query, control_explanations_query,
        num_test_rows, num_control_rows,
        on_columns, min_risk_ratio)

    result = engine.execute(diff_query)
    explanations = []
    for row in result:
        predicates = []
        for column in grouping_set_index[row.grouping_id]:
            if column in on_column_values:
                predicates.append(Predicate(
                    column, Operator.EQUALS, Constant(row[column.name])))
            else:
                # Turn the proxy range bucket column back into a predicate on
                # the range-valued column.
                predicates += test_bucket_predicates[column][row[column.name]]
        explanations.append(Explanation(tuple(predicates), row.risk_ratio))
    result.close()
    return explanations



#!/usr/bin/env python

from pytest import approx
from sqlalchemy.engine import Engine

from datools.models import Column
from datools.models import Table
from datools.table_statistics import column_statistics
from datools.table_statistics import RangeValuedStatistics
from datools.table_statistics import SetValuedStatistics
from .fixtures import generate_synthetic_testdb
from .utils import engine_based_datetime


def test_table_statistics(db_engine: Engine):
    """Tests `column_statistics` on synthetic data in 19 buckets with 9
    values each.
    """
    generate_synthetic_testdb(db_engine)
    statistics = column_statistics(
        db_engine,
        Table('synthetic_data'),
        set())
    assert {
        Column('id'):
            [SetValuedStatistics(171, list(range(1, 101))),
             RangeValuedStatistics([1, 58, 115])],
        Column('same_datetime'):
            [RangeValuedStatistics([
                engine_based_datetime(db_engine,
                                      '2021-05-05 11:00:00.000000')])],
        Column('unique_datetime'):
            [RangeValuedStatistics([
                engine_based_datetime(db_engine,
                                      '2021-05-01 01:00:00.000000'),
                engine_based_datetime(db_engine,
                                      '2021-05-07 04:00:00.000000'),
                engine_based_datetime(db_engine,
                                      '2021-05-13 07:00:00.000000')])],
        Column('bucket_unique_datetime'):
            [RangeValuedStatistics([
                engine_based_datetime(db_engine,
                                      '2021-05-05 01:00:00.000000'),
                engine_based_datetime(db_engine,
                                      '2021-05-05 04:00:00.000000'),
                engine_based_datetime(db_engine,
                                      '2021-05-05 07:00:00.000000')])],
        Column('same_string'):
            [SetValuedStatistics(1, ['hi'])],
        Column('unique_string'):
            [SetValuedStatistics(171,
                                 sorted([f'{bucket}-{row}'
                                         for bucket in range(1, 20)
                                         for row in range(1, 10)])[:100])],
        Column('bucket_unique_string'):
            [SetValuedStatistics(9, [str(x) for x in range(1, 10)])],
        Column('same_float'):
            [RangeValuedStatistics([approx(1.1)])],
        Column('unique_float'):
            [RangeValuedStatistics(
                [approx(1.01), approx(7.04), approx(13.07)])],
        Column('bucket_unique_float'):
            [RangeValuedStatistics([approx(1.0), approx(4.0), approx(7.0)])],
        Column('same_int'):
            [SetValuedStatistics(1, [1]), RangeValuedStatistics([1])],
        Column('unique_int'):
            [SetValuedStatistics(171,
                                 sorted([bucket * 1000 + row
                                         for bucket in range(1, 20)
                                         for row in range(1, 10)])[:100]),
             RangeValuedStatistics([1001, 7004, 13007])],
        Column('bucket_unique_int'):
            [SetValuedStatistics(9, list(range(1, 10))),
             RangeValuedStatistics([1, 4, 7])],
    } == statistics
